'''
    PyBB - Machine, Read/Write, Ip Blocker

    Copyright (c) 2003-2004 Gang Seong Lee <gslee@mail.kw.ac.kr>
    All rights reserved, see COPYING for details.
'''

import os
import time
import re

import cache
import config

class MachineBlockingError(Exception):
    pass
class IpBlockingError(Exception):
    pass

class MachineBlocker:
    def __init__(self):
        self.cache = cache.Cache(50)
    def getIp(self, ip=None):
        return ip and ip or os.environ.get('REMOTE_ADDR', '')
    def checkTime(self, ip=None, timeLimit=0):
        ip = self.getIp(ip)
        try:
            lastTick, items = self.cache.getItem(ip)
        except TypeError:
            lastTick, items = 0, []
        currentTick = time.time()
        self.cache.putItem(ip, (currentTick, items))
        if currentTick - lastTick < timeLimit:
            raise MachineBlockingError, 'time violation %s' % (currentTick-lastTick)
    def clearSequence(self, ip=None):
        ip = self.getIp(ip)
        self.cache.deleteItem(ip)
    def initSequence(self, ip=None):
        ip = self.getIp(ip)
        tick = time.time()
        self.cache.putItem(ip, (tick, []))
        return tick
    def addToSequence(self, ip=None, timeLimit=0, item=None):
        ip = self.getIp(ip)
        try:
            lastTick, items = self.cache.getItem(ip)
        except TypeError:
            lastTick, items = 0, []
        currentTick = time.time()
        if currentTick - lastTick < timeLimit:
            self.cache.putItem(ip, (currentTick, items))
            raise MachineBlockingError, 'time violation %s' % (currentTick-lastTick)
        if items == [] or items and items[-1] != item:
            items.append(item)
        self.cache.putItem(ip, (currentTick, items))
    def compareItems(self, ip=None, timeLimit=0, itemTupleList=()):
        ip = self.getIp(ip)
        lastTick, items = self.cache.getItem(ip)
        currentTick = time.time()
        if currentTick - lastTick < timeLimit:
            self.cache.putItem(ip, (currentTick, items))
            raise MachineBlockingError, 'time violation %s' % (currentTick-lastTick)
        for itemTuple in itemTupleList:
            if tuple(items) == itemTuple:
                break
        else:
            raise MachineBlockingError, 'procedure violation %s' % items
        return True

class RWBlocker:
    def __init__(self, maxReadCount, maxWriteCount, timeInterval):
        self.maxReadCount = maxReadCount
        self.maxWriteCount = maxWriteCount
        self.timeInterval = timeInterval
        self.startTime = time.time() - timeInterval
        if not os.path.exists(config.ipControlFolder):
            os.makedirs(config.ipControlFolder)
        self.ip = os.environ.get('REMOTE_ADDR', '')
        self.read = []
        self.write = []
        self.readFile()
    def readFile(self):
        try:
            f = open(os.path.join(config.ipControlFolder, self.ip))
            for line in f:
                fields = line.split()   # mode(R/W) time
                if len(fields) == 0:
                    continue
                mode, ticks = fields
                if mode == 'R' and ticks > self.startTime:
                    self.read.append(ticks)
                elif mode == 'W' and ticks > self.startTime:
                    self.write.append(ticks)
            self.read.sort()
            self.write.sort()
        except IOError:
            return []
    def writeFile(self):
        f = open(os.path.join(config.ipControlFolder, self.ip), 'w')
        for r in self.read[:self.maxReadCount]:
            f.write('R %s\n' % r)
        for w in self.write[:self.maxWriteCount]:
            f.write('W %s\n' % w)
    def deleteFile(self):
        if os.path.exists(os.path.join(config.ipControlFolder, self.ip)):
            os.remove(os.path.join(config.ipControlFolder, self.ip))
    def appendEvent(self, mode):
        if mode == 'R':
            if self.getReadCount() >= self.maxReadCount:
                raise MachineBlockingError, 'max read count exceeded %s' % self.getReadCount()
            self.read.append(time.time())
        elif mode == 'W':
            if self.getWriteCount() >= self.maxWriteCount:
                raise MachineBlockingError, 'max write count exceeded %s' % self.getWriteCount()
            self.write.append(time.time())
        self.writeFile()
    def getReadCount(self):
        return len(self.read)
    def getWriteCount(self):
        return len(self.write)


class IpController:
    def __init__(self):
        self.setAllowIpsByString(config.allowIpList)
        self.setDenyIpsByString(config.allowIpList)

    def setAllowIpsByString(self, allowIpString):
        self.allowIpList = filter(None, allowIpString.split())
        return self.allowIpList

    def setDenyIpsByString(self, denyIpString):
        self.denyIpList = filter(None, denyIpString.split())
        return self.denyIpList
    
    def ipCompare(self, userIp, comparedIp):
        splitedUserIp = userIp.split('.')
        splitedComparedIp = filter(lambda byte: byte.strip(), comparedIp.split('.') )
        for index in range(len(splitedComparedIp)):
            byteOfUserIp = splitedUserIp[index].zfill(3)
            byteOfComparedIp = splitedComparedIp[index]
            if byteOfComparedIp.find('*') < 0: 
                byteOfComparedIp = byteOfComparedIp.zfill(3)
            else:
                byteOfComparedIp = re.sub('\*', '*', byteOfComparedIp)
                space = 3-len(byteOfComparedIp)
                byteOfComparedIp = byteOfComparedIp.replace('*', '?' * (space+1))
            byteOfComparedIp = byteOfComparedIp.replace('?', '.?')
            if not re.match(byteOfComparedIp, byteOfUserIp):
                break
        else:
            return True
        return False
 
    def inRange(self, userIp, ipList):
        for comparedIp in ipList:
            if self.ipCompare(userIp, comparedIp):
                return True
        return False

    def checkIpAdmission(self, userIp):
        if self.allowIpList and not self.denyIpList:
            ret = self.inRange(userIp, self.allowIpList)
            if ret: return ret
            raise IpBlockingError, userIp
        if self.allowIpList and self.inRange(userIp, self.allowIpList):
            return True
        if self.inRange(userIp, self.denyIpList):
            raise IpBlockingError, userIp
        return True # complement area